"""

Copyright (C) 2017-2021 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from sregistry.logger import bot
from sregistry.auth import basic_auth_header
from sregistry.main import ApiConnection
import json
import sys
import os

# The core of nvidia is actually docker
from sregistry.main.docker.api import (
    create_metadata_tar,
    download_layers,
    get_manifest_selfLink,
    get_config,
    get_digests,
    get_environment_tar,
    get_layer,
    get_layerLink,
    get_manifest,
    get_manifests,
    get_download_cache,
    get_size,
    extract_env,
    extract_labels,
    extract_runscript,
    update_token,
)

from sregistry.main.docker.pull import _pull
from .pull import pull


class Client(ApiConnection):
    def __init__(self, secrets=None, **kwargs):
        """to work with nvidia container registry, we do the following:

        1. set the base to index.docker.io, or defined in environment
        2. assume starting with v2 schema, we won't reverse layers
        3. set headers to ask for version 2 schema first
        4. update secrets, 1st priority environment, then .docker/config.json
        5. update headers based on secrets

        """
        self._set_base()
        self.reverseLayers = False
        self._reset_headers()
        self._update_secrets()
        self._update_headers()
        super(Client, self).__init__(**kwargs)

    def _reset_headers(self):
        """reset headers is called from update_headers, and will update the
        headers based on what is found with the client secrets
        """

        # specify wanting version 2 schema
        # meaning the correct order of digests
        # returned (base to child)
        accept = "application/vnd.docker.distribution.manifest.v2+json,"
        accept += "application/vnd.docker.distribution.manifest.list.v2+json,"
        accept += "application/json"

        self.headers = {
            "Accept": accept,
            "Content-Type": "application/json; charset=utf-8",
        }

    def _set_base(self):
        """set the API base or default to use Docker Hub. The user is able
        to set the base, api version, and protocol via a settings file
        of environment variables:

        SREGISTRY_NVIDIA_BASE: defaults to nvcr.io
        SREGISTRY_NVIDIA_TOKEN: defaults to $oauthtoken
        SREGISTRY_NVIDIA_VERSION: defaults to v2
        SREGISTRY_NVIDIA_NO_HTTPS: defaults to not set (so https)

        """
        base = self._get_setting("SREGISTRY_NVIDIA_BASE")
        version = self._get_setting("SREGISTRY_NVIDIA_VERSION")

        if base is None:
            base = "nvcr.io"

        if version is None:
            version = "v2"

        nohttps = self._get_setting("SREGISTRY_NVIDIA_NOHTTPS")
        if nohttps is None:
            nohttps = "https://"
        else:
            nohttps = "http://"

        # <protocol>://<base>/<version>
        self.base = "%s%s/%s" % (nohttps, base.strip("/"), version)

    def _update_secrets(self):
        """update secrets will take a secrets credential file
        either located at .sregistry or the environment variable
        SREGISTRY_CLIENT_SECRETS and update the current client
        secrets as well as the associated API base. For the case of
        using Docker Hub, if we find a .docker secrets file, we update
        from there.
        """
        # If the user has defined secrets, use them
        token = self._required_get_and_update("SREGISTRY_NVIDIA_TOKEN")
        username = self._get_and_update_setting("SREGISTRY_NVIDIA_USERNAME")

        if username is None:
            username = "$oauthtoken"

        # Option 1: the user exports username and password
        if token is not None:
            auth = basic_auth_header(username, token)
            self.headers.update(auth)

    def __str__(self):
        return type(self)


# Functions exposed to the client
Client.pull = pull
Client._pull = _pull

# Api functions for image layers and manifests (hidden)
Client._create_metadata_tar = create_metadata_tar
Client._download_layers = download_layers
Client._extract_runscript = extract_runscript
Client._extract_labels = extract_labels
Client._extract_env = extract_env
Client._get_config = get_config
Client._get_digests = get_digests
Client._get_download_cache = get_download_cache
Client._get_layer = get_layer
Client._get_layerLink = get_layerLink
Client._get_manifest = get_manifest
Client._get_manifests = get_manifests
Client._get_size = get_size
Client._update_token = update_token
Client._get_manifest_selfLink = get_manifest_selfLink
Client._get_environment_tar = get_environment_tar



"""

Copyright (C) 2017-2021 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from sregistry.logger import bot


def search(self, query=None, args=None):
    """query a bucket for images that match a particular pattern. If no
    query is provided, all images in the bucket are listed that have type
    "container" in the metadata and client "sregistry"
    """

    if query is not None:
        return self._container_query(query)
    return self._search_all()


def list_containers(self):
    """return a list of containers, determined by finding the metadata field
    "type" with value "container." We alert the user to no containers
    if results is empty, and exit

    {'metadata': {'items':
                           [
                            {'key': 'type', 'value': 'container'}, ...
                           ]
                 }
    }

    """
    results = []
    for image in self._bucket.list_blobs():
        if image.metadata is not None:
            if "type" in image.metadata:
                if image.metadata["type"] == "container":
                    results.append(image)

    if len(results) == 0:
        bot.info("No containers found, based on metadata type:container")

    return results


def search_all(self):
    """a "list all" search that doesn't require a query. Here we return to
    the user all objects that have custom metadata value of "container"

    IMPORTANT: the upload function adds this metadata. For a container to
    be found by the client, it must have the type as container in metadata.
    """

    results = self._list_containers()

    bot.info("[gs://%s] Containers" % self._bucket_name)

    rows = []
    for i in results:
        size = round(i.size / (1024 * 1024.0))
        size = ("%s MB" % size).rjust(10)
        rows.append([size, i.metadata["uri"]])

    bot.table(rows)
    return rows


def container_query(self, query, quiet=False):
    """search for a specific container.
    This function would likely be similar to the above, but have different
    filter criteria from the user (based on the query)
    """
    results = self._list_containers()

    matches = []
    for result in results:
        for _, val in result.metadata.items():
            if query in val and result not in matches:
                matches.append(result)
            elif query in result.name and result not in matches:
                matches.append(result)

    if not quiet:
        bot.info("[gs://%s] Found %s containers" % (self._bucket_name, len(matches)))
        for image in matches:
            size = round(image.size / (1024 * 1024.0))
            bot.custom(prefix=image.name, color="CYAN")
            bot.custom(prefix="id:     ", message=image.id)
            bot.custom(prefix="uri:    ", message=image.metadata["uri"])
            bot.custom(prefix="updated:", message=image.updated)
            bot.custom(prefix="size:  ", message=" %s MB" % (size))
            bot.custom(prefix="md5:    ", message=image.md5_hash)
            bot.newline()
    return matches



"""

Copyright (C) 2018-2021 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from sregistry.defaults import DISABLE_SSL_CHECK
from sregistry.logger import bot

from requests.exceptions import HTTPError
import requests

import json
import os
import re
import shutil
import sys
import tempfile


################################################################################
## Shared Tasks for the Worker
################################################################################


def download_task(url, headers, destination, download_type="layer"):
    """download an image layer (.tar.gz) to a specified download folder.
    This task is done by using local versions of the same download functions
    that are used for the client.
    core stream/download functions of the parent client.

    Parameters
    ==========
    image_id: the shasum id of the layer, already determined to not exist
    repo_name: the image name (library/ubuntu) to retrieve
    download_folder: download to this folder. If not set, uses temp.


    """
    # Update the user what we are doing
    bot.verbose("Downloading %s from %s" % (download_type, url))

    # Step 1: Download the layer atomically
    file_name = "%s.%s" % (destination, next(tempfile._get_candidate_names()))

    tar_download = download(url, file_name, headers=headers)

    try:
        shutil.move(tar_download, destination)
    except:
        msg = "Cannot untar layer %s," % tar_download
        msg += " was there a problem with download?"
        bot.exit(msg)

    return destination


################################################################################
## Base Functions for Tasks
##
##  These basic tasks are intended for the worker to use, without needing
##  to pickle them for multiprocessing. It works because they don't belong
##  to a client (which we cannot pickle) and are imported by the worker
##  functions directly.
##
################################################################################


def post(url, headers=None, data=None, return_json=True):
    """post will use requests to get a particular url"""
    bot.debug("POST %s" % url)
    return call(
        url, headers=headers, func=requests.post, data=data, return_json=return_json
    )


def get(url, headers=None, token=None, data=None, return_json=True):
    """get will use requests to get a particular url"""
    bot.debug("GET %s" % url)
    return call(
        url, headers=headers, func=requests.get, data=data, return_json=return_json
    )


def download(url, file_name, headers=None, show_progress=True):
    """stream to a temporary file, rename on successful completion

    Parameters
    ==========
    file_name: the file name to stream to
    url: the url to stream from
    headers: additional headers to add
    """

    fd, tmp_file = tempfile.mkstemp(prefix=("%s.tmp." % file_name))
    os.close(fd)

    if DISABLE_SSL_CHECK is True:
        bot.warning("Verify of certificates disabled! ::TESTING USE ONLY::")

    verify = not DISABLE_SSL_CHECK

    # Does the url being requested exist?
    if requests.head(url, verify=verify).status_code in [200, 401]:
        response = stream(url, headers=headers, stream_to=tmp_file)

        if isinstance(response, HTTPError):
            bot.exit("Error downloading %s, exiting." % url)

        shutil.move(tmp_file, file_name)
    else:
        bot.error("Invalid url or permissions %s" % url)
    return file_name


def stream(url, headers, stream_to=None, retry=True):
    """stream is a get that will stream to file_name. Since this is a worker
    task, it differs from the client provided version in that it requires
    headers.
    """

    bot.debug("GET %s" % url)

    if DISABLE_SSL_CHECK is True:
        bot.warning("Verify of certificates disabled! ::TESTING USE ONLY::")

    # Ensure headers are present, update if not
    response = requests.get(
        url, headers=headers, verify=not DISABLE_SSL_CHECK, stream=True
    )

    # Deal with token if necessary
    if response.status_code == 401 and retry is True:
        headers = update_token(response, headers)
        return stream(url, headers, stream_to, retry=False)

    if response.status_code == 200:

        # Keep user updated with Progress Bar
        content_size = None
        if "Content-Length" in response.headers:
            progress = 0
            content_size = int(response.headers["Content-Length"])
            bot.show_progress(progress, content_size, length=35)

        chunk_size = 1 << 20
        with open(stream_to, "wb") as filey:
            for chunk in response.iter_content(chunk_size=chunk_size):
                filey.write(chunk)
                if content_size is not None:
                    progress += chunk_size
                    bot.show_progress(
                        iteration=progress,
                        total=content_size,
                        length=35,
                        carriage_return=False,
                    )

        # Newline to finish download
        sys.stdout.write("\n")

        return stream_to

    bot.exit("Problem with stream, response %s" % response.status_code)


def call(
    url, func, data=None, headers=None, return_json=True, stream=False, retry=True
):

    """call will issue the call, and issue a refresh token
    given a 401 response, and if the client has a _update_token function

    Parameters
    ==========
    func: the function (eg, post, get) to call
    url: the url to send file to
    headers: headers for the request
    data: additional data to add to the request
    return_json: return json if successful
    """

    if DISABLE_SSL_CHECK is True:
        bot.warning("Verify of certificates disabled! ::TESTING USE ONLY::")

    if data is not None:
        if not isinstance(data, dict):
            data = json.dumps(data)

    response = func(
        url=url, headers=headers, data=data, verify=not DISABLE_SSL_CHECK, stream=stream
    )

    # Errored response, try again with refresh
    if response.status_code in [500, 502]:
        bot.exit("Beep boop! %s: %s" % (response.reason, response.status_code))

    # Errored response, try again with refresh
    if response.status_code == 404:
        bot.exit("Beep boop! %s: %s" % (response.reason, response.status_code))

    # Errored response, try again with refresh
    if response.status_code == 401:

        # If client has method to update token, try it once
        if retry is True:

            # A result of None indicates no update to the call
            headers = update_token(response, headers)
            return call(
                url,
                func,
                data=data,
                headers=headers,
                return_json=return_json,
                stream=stream,
                retry=False,
            )

        bot.exit(
            "Your credentials are expired! %s: %s"
            % (response.reason, response.status_code)
        )

    elif response.status_code == 200:

        if return_json:

            try:
                response = response.json()
            except ValueError:
                bot.exit("The server returned a malformed response.")

    return response


def update_token(response, headers):
    """update_token uses HTTP basic authentication to attempt to authenticate
    given a 401 response. We take as input previous headers, and update
    them.

    Parameters
    ==========
    response: the http request response to parse for the challenge.

    """

    not_asking_auth = "Www-Authenticate" not in response.headers
    if response.status_code != 401 or not_asking_auth:
        bot.exit("Authentication error, exiting.")

    challenge = response.headers["Www-Authenticate"]
    regexp = '^Bearer\s+realm="(.+)",service="(.+)",scope="(.+)",?'
    match = re.match(regexp, challenge)

    if not match:
        bot.exit("Unrecognized authentication challenge, exiting.")

    realm = match.group(1)
    service = match.group(2)
    scope = match.group(3).split(",")[0]

    token_url = realm + "?service=" + service + "&expires_in=900&scope=" + scope

    response = get(token_url)

    try:
        token = response["token"]
        token = {"Authorization": "Bearer %s" % token}
        headers.update(token)

    except:
        bot.exit("Error getting token.")

    return headers



"""

Copyright (C) 2017-2021 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from sregistry.logger import bot


def main(args, parser, extra):

    from sregistry.main import get_client

    cli = get_client(quiet=args.quiet)
    cli.announce(args.command)

    # If the client doesn't have the command, exit
    if not hasattr(cli, "mv"):
        msg = "move is not implemented for %s. Why don't you add it?"
        bot.exit(msg % cli.client_name)

    cli.mv(image_name=args.name, path=args.path)
